#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"This module define the basic data structure that are used in this library"

from itertools import permutations, combinations

import numpy as np
import scipy as sp

import fastdtw

from eeglib.features import (bandPower, hjorthActivity, hjorthMobility,
                             hjorthComplexity, sampEn, LZC, DFA, HFD, PFD,
                             synchronizationLikelihood)
from eeglib.preprocessing import bandPassFilter
from eeglib.auxFunctions import listType

# Default bands ranges
defaultBands = {"delta": (1, 4), "theta": (4, 7),
                "alpha": (8, 12), "beta": (12, 30)}


class SampleWindow:
    """
    This class is a data structure that stores the signal data and works like a
    sliding window.

    Attributes
    ----------
    windowSize: int
        The maximun samples the window will store.
    channelNumber: int
        The number of channels of samples the window will handle.
    window: list of lists
        It stores the data.
    """

    def __init__(self, windowSize, channelNumber,names=None):
        """
        Creates a SampleWindow wich stores the value of windowSize as the
        number of samples and the value of channelNumber as the number of
        channels.

        Parameters
        ----------
        windowSize: int
            The size of the sliding window.

        channelNumber: int
            The number of channels recording simultaneously.
        names: list of strings, optional
            The optional names that can be used to refer to each channel.
        """
        self.windowSize = windowSize
        self.channelNumber = channelNumber

        self.window = np.zeros((channelNumber,windowSize))
        if names:
            self._names=names
            self._windowDict={n:l for (n,l) in zip(names,self.window)}
        else:
            self._names=None

    def set(self, samples, columnMode=False):
        """
        Sets multiple samples at a time. The sample number must be the same as
        the window size.

        Parameters
        ----------
        samples: array_like of 2 dimensions
            Samples of data with a size equals to window.

        columnMode: boolean, optional
            By default it is assummed that the shape of the data given is
            nSamples X nchannels. If the given data is the inverse,
            columnMode should be True.
        """
        if hasattr(samples, "__getitem__") \
            and hasattr(samples[0], "__getitem__"):
            if((len(samples) == self.windowSize and
                len(samples[0]) == self.channelNumber and not columnMode)
                or (len(samples) == self.channelNumber and
                len(samples[0]) == self.windowSize and columnMode)):
                if not columnMode:
                    self.window = np.transpose(np.array(samples))
                else:
                    self.window = np.array(samples)
                if self._names:
                    for i in range(self.channelNumber):
                        self._windowDict[self._names[i]]=self.window[i]
            else:
                raise ValueError(("The number of samples must be equal to the "
                                 + "window size and each sample length must be"
                                 +" a equals to channelNumber ( {0} ) if not "
                                 +"in columnMode and viceversa if in "+
                                 "columnMode.").format(self.channelNumber))
        else:
            raise ValueError("Samples must be a subscriptable object wich "+
                             "contains subcriptable objects.")

    def getIndicesList(self, i):
        """
        Returns a list of numeric indices from a combined type of indices.
        """
        if i is None:
            return [*range(0, self.channelNumber)]
        if isinstance(i, slice):
            return self.getIndicesList(None)[i]
        if np.issubdtype(type(i), np.integer):
            return [i]
        if isinstance(i, str):
            return [self._names.index(i)]
        if isinstance(i, list):
            return sum((self.getIndicesList(j) for j in i), start=[])
        raise ValueError("This method only accepts int, str, list or" +
                             "slice types and not %s"%type(i))

    def getChannel(self, i=None):
        """
        Returns an array containing the data of the the selected channel/s.

        Parameters
        ----------
        i: Variable type, optional
            * int:  the index of the channel.
            * str:  the name of the channel.
            * list of strings and integers:  a list of channels that will be
              returned as a 2D ndarray.
            * slice:  a slice selecting the range of channels that wll be
              returned as a 2D ndarray.
            * None: all the data returned as a 2D ndarray

        Returns
        -------
        numpy.ndarray
            Can be a one or a two dimension matrix, depending of the
            parameters.
        """
        if i is None:
            return self.window

        if isinstance(i, str):
            if self._names:
                return self._windowDict[i]
            raise ValueError("There aren't names asociated to the channels.")

        if isinstance(i, list):
            selectedChannels=np.zeros((len(i),self.windowSize),dtype=float)
            for c, e in enumerate(i):
                if type(e) in [int, str]:
                    selectedChannels[c]=self.getChannel(e)
                else:
                    raise ValueError("The list can only contain int and/or " +
                                     "strings.")
            return selectedChannels

        if type(i) in [int,slice]:
            return self.window[i]

        raise ValueError("This method only accepts int, str, list or " +
                             "slice types and not %s"%type(i))

    def getPairsIndicesList(self, i, allPermutations=False):
        """
        Returns a list of tuples of numeric indices from a combined type of
        indices.
        """
        permutate = permutations if allPermutations else combinations
        if i is None:
            return permutate([*range(0, self.channelNumber)], 2)

        if isinstance(i, slice):
            return self.getPairsIndicesList(self.getIndicesList(None)[i])

        if isinstance(i, tuple) and len(i) == 2:
            return [tuple(self.getIndicesList(list(i)))]

        if isinstance(i, list):
            if listType(i) is tuple:
                return [self.getPairsIndicesList(c)[0] for c in i]
            return permutate(self.getIndicesList(i), 2)

        raise ValueError("This method only accepts int, str, list or " +
                             "slice types and not %s"%type(i))

    def getPairOfChannels(self, channels = None, allPermutations=False):
        """
        Applies a function that uses two signals by selecting the channels to
        use. It will apply the function to different channels depending on the
        parameters. Note: a single channel can be selected by using an int or
        a string if a name for the channel was specified.

        Parameters
        ----------
        channels: Variable type, optional
            * tuple of lenght 2 containing channel indexes: applies the
              function to the two channels specified by the tuple.
            * list of tuples(same restrictions than the above): applies the
              function to every tuple in the list.
            * list of index: creates combinations of channels specifies in the
              list depending of the allPermutations parameter.
            * None: Are channels are used in the same way than in the list
              above. This is the default value.

        allPermutations: bool, optional
            Only used when channels is a list of index or None. If True all
            permutations of the channels in every order are used; if False only
            combinations of different channels are used. Example: with the list
            [0, 2, 4] and allPermutations = True, the channels used will be
            (0,2), (0,4), (2,0), (2,4), (4,0), (4,2); meanwhile, with
            allPermutations = False, the channels used will be: (0,2), (0,4),
            (2,4). Default: False.

        Returns
        -------
        If a tuple is passed the it returns the result of applying the function
        to the channels specified in the tuple. If another valid value is
        passed, the method returns a dictionary, being the key the two channels
        used and the value the result of applying the function to those
        channels.
        """
        permutate = permutations if allPermutations else combinations

        if channels is None:
            channels = [*range(self.channelNumber)]
        elif isinstance(channels, slice):
            channels = [*range(self.channelNumber)[channels]]

        if isinstance(channels, tuple):
            if len(channels) != 2:
                raise ValueError("If you specify a tuple, it must have " +
                                 "exactly two(2) elements")
            return (self.getChannel(channels[0]), self.getChannel(channels[1]))

        if isinstance(channels, list):
            t = listType(channels)
            if t is tuple:
                acs = np.array(channels)
                if len(acs.shape) == 2 and acs.shape[1] == 2:
                    return [(self.getChannel(c[0]), self.getChannel(c[1]))
                            for c in channels]
                raise ValueError("The tuples of the list must have " +
                                 "exactly two(2) elements")
            try:
                return permutate(self.getChannel(channels), 2)
            except ValueError as err:
                raise ValueError("The list must contain either tuples " +
                                 "or integers and strings, but no "+
                                 "combinations of them.") from err

        raise ValueError("This method only accepts list (either of int "+
                         "and str or tuples) or tuple types and not %s" %
                         type(channels))

    def __getitem__(self, n):
        return self.window[:,n]

    # Returns a string representation of the inner list
    def __str__(self):
        return str(self.window)


class EEG:
    """
    This class apply sginal analysis functions over the data stored in its
    attribute window.

    Attributes
    ----------
    windowSize: int
        The maximun samples the window will store.
    sampleRate: int
        The number of samples per second
    channelNumber: int
        The number of channels of samples the window will handle.
    window: SampleWindow
        It stores the data.
    """

    def __init__(self, windowSize, sampleRate, channelNumber, names=None):
        """
        Parameters
        ----------
        windowSize: int
            The maximun samples the window will store.
        sampleRate: int
            The number of samples per second
        channelNumber: int
            The number of channels of samples the window will handle.
        names: list of strings, optional
            The optional names that can be used to refer to each channel.
        """
        self.windowSize = windowSize
        self.sampleRate = sampleRate
        self.channelNumber = channelNumber
        self.window = SampleWindow(windowSize, channelNumber,names=names)
        self.outputMode = "array"

    # Function to handle the windowFunction parameter
    def _handleWindowFunction(self, windowFunction):
        if windowFunction is None:
            return 1

        if type(windowFunction) in (str, tuple):
            return sp.signal.get_window(windowFunction, self.windowSize)

        if isinstance(windowFunction, np.ndarray):
            if len(windowFunction) == self.windowSize:
                return windowFunction
            raise ValueError("The size of windowFunction is not the same " +
                             "as windowSize.")

        raise ValueError("Not a valid type for windowFunction.")


    def set(self, samples, columnMode=False):
        """
        Sets multiple samples at a time into the window. The sample number must
        be the same as the window size.

        Parameters
        ----------
        samples: array_like of 2 dimensions
            Samples of data with a size equals to window.

        columnMode: boolean, optional
            By default it is assummed that the shape of the data given is
            nSamples X nchannels. If the given data is the inverse,
            columnMode should be True.
        """
        self.window.set(samples, columnMode)

    def getChannel(self, i=None):
        """
        Returns the raw data stored at the given index of the windows.

        Parameters
        ----------
        i: Variable type, optional
            * int:  the index of the channel.
            * str:  the name of the channel.
            * list of strings and integers:  a list of channels that will be
              returned as a 2D ndarray.
            * slice:  a slice selecting the range of channels that wll be
              returned as a 2D ndarray.
            * None: all the data returned as a 2D ndarray

        Returns
        -------
        list
            The list of values of a specific channel.
        """
        return self.window.getChannel(i)

    def _applyFunctionTo(self, function, i=None):
        """
        Returns the raw data stored at the given index of the windows.

        Parameters
        ----------
        i: Variable type, optional
            * int:  the index of the channel.
            * str:  the name of the channel.
            * list of strings and integers:  a list of channels that will be
              returned as a 2D ndarray.
            * slice:  a slice selecting the range of channels that wll be
              returned as a 2D ndarray.
            * None: all the data returned as a 2D ndarray

        Returns
        -------
        list
            The list of values of a specific channel.
        """
        data=self.window.getChannel(i)
        if len(np.shape(data))==1:
            return function(data)

        if self.outputMode=="dict":
            indices = self.window.getIndicesList(i)
            if self.window._names:
                keys = np.array(self.window._names)[indices]
            else:
                keys=indices
            return {key:function(d) for key, d in zip(keys, data)}

        return np.array([function(d) for d in data])

    def _applyFunctionTo2C(self, function, channels = None,
                            allPermutations=False):
        """
        Applies a function that uses two signals by selecting the channels to
        use. It will apply the function to different channels depending on the
        parameters. Note: a single channel can be selected by using an int or
        a string if a name for the channel was specified.

        Parameters
        ----------
        function: function
            A function that receives two streams of data and will be applied to
            the selected channels.

        channels: Variable type, optional
            * tuple of lenght 2 containing channel indexes: applies the
              function to the two channels specified by the tuple.
            * list of tuples(same restrictions than the above): applies the
              function to every tuple in the list.
            * list of index: creates combinations of channels specifies in the
              list depending of the allPermutations parameter.
            * None: Are channels are used in the same way than in the list
              above. This is the default value.

        allPermutations: bool, optional
            Only used when channels is a list of index or None. If True all
            permutations of the channels in every order are used; if False only
            combinations of different channels are used. Example: with the list
            [0, 2, 4] and allPermutations = True, the channels used will be
            (0,2), (0,4), (2,0), (2,4), (4,0), (4,2); meanwhile, with
            allPermutations = False, the channels used will be: (0,2), (0,4),
            (2,4). Default: False.

        Returns
        -------
        If a tuple is passed the it returns the result of applying the function
        to the channels specified in the tuple. If another valid value is
        passed, the method returns a dictionary, being the key the two channels
        used and the value the result of applying the function to those
        channels.
        """
        data = self.window.getPairOfChannels(channels, allPermutations)

        if isinstance(data, tuple):
            return function(*data)

        if self.outputMode == "dict":
            indices = self.window.getPairsIndicesList(channels,allPermutations)
            if self.window._names:
                keys = [(self.window._names[i[0]], self.window._names[i[1]])
                        for i in indices]
            else:
                keys=indices
            return {key:function(*value) for key, value in zip(keys, data)}

        return [function(*value) for value in data]

    def DFT(self, i=None, windowFunction=None, output="complex",
            onlyPositiveFrequencies = False):
        """
        Returns the Discrete Fourier Transform of the data at a given index of
        the window.

        Parameters
        ----------
        i: Variable type, optional
            * int:  the index of the channel.
            * str:  the name of the channel.
            * list of strings and integers:  a list of channels that will be
              used.
            * slice: a slice selecting the range of channels that will be
              used.
            * None: all the channels will be used.

        windowFunction: str, tuple or numpy.ndarray, optional
            This can be a string with the name of the function, a tuple with a
            str with the name of the funcion in the first position and the
            parameters of the funcion in the nexts or a numpy array with a size
            equals to the window size. In the first case an array with the
            size of windowSize will be created. The created array will be
            multiplied by the data in the window before FFT is used.

        output: str, optional
            * "complex": default output of the FFT, x+yi
            * "magnitude": computes the magnitude of the FFT, sqrt(x^2+y^2)
            * "phase": computes the phase of the FFT, atan2(yi,x)

        Returns
        -------
        numpy.ndarray
            An array with the result of the Fourier Transform. If more than one
            channel was selected the array will be of 2 Dimensions.
        """
        windowFunction=self._handleWindowFunction(windowFunction)
        f1 = lambda x:np.fft.fft(x*windowFunction)
        output = output.lower()
        if output == "magnitude":
            f = lambda x: np.abs(f1(x))
        elif output == "phase":
            f = lambda x: np.angle(f1(x))
        elif output == "complex":
            f=f1
        else:
            raise ValueError('Only "complex", "magnitude" or "phase" are ' +
                             'valid values por output parameter')

        if onlyPositiveFrequencies:
            auxF = f
            f = lambda x: auxF(x)[:self.windowSize//2+1]

        return self._applyFunctionTo(f, i)


    def PSD(self, i = None, windowFunction = "hann", nperseg = None,
            retFrequencies = False):
        """
        Returns the Power Spectral Density of the data at a given index of
        the window using the Welch method.

        Parameters
        ----------
        i: Variable type, optional
            * int:  the index of the channel.
            * str:  the name of the channel.
            * list of strings and integers:  a list of channels that will be
              used.
            * slice: a slice selecting the range of channels that will be
              used.
            * None: all the channels will be used.

        windowFunction: str or tuple or array_like, optional
            Desired window to use. If window is a string or tuple, it is passed
            to get_window to generate the window values, which are DFT-even by
            default. See get_window for a list of windows and required
            parameters. If window is array_like it will be used directly as the
            window and its length must be nperseg. Defaults to a Hann window.

        nperseg: int, optional
            Length of each segment. Defaults to None, but if window is str or
            tuple, is set to 256, and if window is array_like, is set to the
            length of the window.

        retFrequencies: bool, optional
            If True two arrays will be raturned for each channel, one with the
            frequencies and another with the spectral density of those
            frequencies.

        Returns
        -------
        numpy.ndarray
            An array with the result of the Fourier Transform. If more than one
            channel was selected the array will be of 2 Dimensions.
        """
        if nperseg is None:
            nperseg = self.windowSize//2
        f1 = lambda x: sp.signal.welch(x, self.sampleRate,
                                       window = windowFunction,
                                       nperseg = nperseg)
        if not retFrequencies:
            f = lambda x: f1(x)[1]
        else:
            f = f1
        return self._applyFunctionTo(f, i)


    def bandPower(self, i=None, bands=defaultBands, spectrumFrom = "DFT",
                  windowFunction="hann", nperseg = None, normalize=False):
        """
        Returns the power of each band at the given index.

        Parameters
        ----------
        i: Variable type, optional
            * int:  the index of the channel.
            * str:  the name of the channel.
            * list of strings and integers:  a list of channels that will be
              used.
            * slice: a slice selecting the range of channels that will be
              used.
            * None: all the channels will be used.

        bands: dict, optional
            This parameter is used to indicate the bands that are going to be
            used. It is a dict with the name of each band as key and a tuple
            with the lower and upper bounds as value.

        spectrumFrom: str, optional
            * "DFT": uses the spectrum from the DFT of the signal.
            * "PSD": uses the spectrum from the PSD of the signal.

        windowFunction: str or tuple or array_like, optional
            Desired window to use. If window is a string or tuple, it is passed
            to get_window to generate the window values, which are DFT-even by
            default. See get_window for a list of windows and required
            parameters. If window is array_like it will be used directly as the
            window and its length must be nperseg. Defaults to a Hann window.

        nperseg: int, optional
            This parameter is only relevant when powerFrom is "PSD", else it is
            ignored. Length of each segment. Defaults to None, but if window is
            str or tuple, is set to 256, and if window is array_like, is set to
            the length of the window.

        normalize: bool, optional
            If True the each band power is divided by the total power of the
            spectrum. Default False.

        Returns
        -------
        dict or list of dict
            The keys are the name of each band and the values are the mean of
            the magnitudes. If more than one channel was selected the return
            object will be a list containing the dict for each channel selected
        """
        bands={key:self.getBoundsForBand(b) for key,b in bands.items()}
        freqRes = self.sampleRate/self.windowSize
        if spectrumFrom.upper() == "DFT":
            spectrum = self.DFT(i, windowFunction, output="magnitude",
                                onlyPositiveFrequencies=True)
        elif spectrumFrom.upper() == "PSD":
            if nperseg is None:
                nperseg = self.windowSize//2
            spectrum = self.PSD(i, windowFunction, nperseg)
            factor = self.windowSize//nperseg
            bands = {k:(b[0]//factor, b[1]//factor) for k, b in bands.items()}
            freqRes*=factor
        else:
            raise ValueError('%s is not a valid value for '%spectrumFrom +
                             'powerFrom parameter. Only "DFT" or "PSD" are ' +
                             'valid values.')

        if isinstance(spectrum, dict):
            return {key:bandPower(s, bands, freqRes, normalize) for key, s
                    in spectrum.items()}

        if spectrum.ndim==1:
            return bandPower(spectrum, bands, freqRes, normalize)

        return [bandPower(s, bands, freqRes, normalize) for s in spectrum]

    def getBoundsForBand(self, bandBounds):
        """
        Returns the bounds of each band depending of the sample rate and the
        window size.

        Parameters
        ----------
        bandbounds: tuple
            A tuple containig the lower and upper bounds of a frequency band.

        Returns
        -------
        tuple
            A tuple containig the the new bounds of the given band.
        """
        return tuple(map(lambda val:
            int(val * self.windowSize / self.sampleRate), bandBounds))

    def getSignalAtBands(self, i=None, bands=defaultBands):
        """
        Rebuilds the signal from a component i but only in the specified
        frequency bands.

        Parameters
        ----------
        i: Variable type, optional
            * int:  the index of the channel.
            * str:  the name of the channel.
            * list of strings and integers:  a list of channels that will be
              used.
            * slice: a slice selecting the range of channels that will be
              used.
            * None: all the channels will be used.

        bands: dict, optional
            This parameter is used to indicate the bands that are going to be
            used. It is a dict with the name of each band as key and a tuple
            with the lower and upper bounds as value.

        Returns
        -------
        dict of numpy.ndarray (1D or 2D)
            The keys are the same keys the bands dictionary is using. The
            values are the signal filtered in every band at the given index of
            the window. If more than one channel is selected the return object
            will be a dict containing 2D arrays in which each row is a signal
            filtered at the corresponding channel.
        """

        bandsSignals = {}
        for key, band in bands.items():
            bounds = self.getBoundsForBand(band)
            bandsSignals[key]=self._applyFunctionTo(lambda x:
                bandPassFilter(x, self.sampleRate, bounds[0], bounds[1]), i)

        return bandsSignals

    def PFD(self, i=None):
        """
        Returns the Petrosian Fractal Dimension at the given index of the
        window.

        Parameters
        ----------
        i: Variable type, optional
            * int:  the index of the channel.
            * str:  the name of the channel.
            * list of strings and integers:  a list of channels that will be
              used.
            * slice: a slice selecting the range of channels that will be
              used.
            * None: all the channels will be used.

        Returns
        -------
        float or array
            The resulting value. If more than one channe was selected the
            return object will be a 1D array containing the result of the
            procesing.
        """
        return self._applyFunctionTo(PFD,i)

    def HFD(self, i=None, kMax=None):
        """
        Returns the Higuchi Fractal Dimension at the given index of the window.

        Parameters
        ----------
        i: Variable type, optional
            * int:  the index of the channel.
            * str:  the name of the channel.
            * list of strings and integers:  a list of channels that will be
              used.
            * slice: a slice selecting the range of channels that will be
              used.
            * None: all the channels will be used.
        kmax: int, optional
            By default it will be windowSize//2.

        Returns
        -------
        float or array
            The resulting value. If more than one channe was selected the
            return object will be a 1D array containing the result of the
            procesing.
        """
        return self._applyFunctionTo(lambda v:HFD(v,kMax),i)

    # Hjorth Parameters:
    def hjorthActivity(self, i=None):
        """
        Returns the Hjorth Activity at the given channel

        Parameters
        ----------
        i: int or string, optional
            Index or name of the channel.

        Returns
        -------
        float or array
            The resulting value. If more than one channe was selected the
            return object will be a 1D array containing the result of the
            procesing.
        """
        return self._applyFunctionTo(hjorthActivity,i)

    def hjorthMobility(self, i=None):
        """
        Returns the Hjorth Mobility at the given channel

        Parameters
        ----------
        i: Variable type, optional
            * int:  the index of the channel.
            * str:  the name of the channel.
            * list of strings and integers:  a list of channels that will be
              used.
            * slice: a slice selecting the range of channels that will be
              used.
            * None: all the channels will be used.

        Returns
        -------
        float or array
            The resulting value. If more than one channe was selected the
            return object will be a 1D array containing the result of the
            procesing.
        """
        return self._applyFunctionTo(hjorthMobility,i)

    def hjorthComplexity(self, i=None):
        """
        Returns the Hjorth Complexity at the given channel

        Parameters
        ----------
        i: Variable type, optional
            * int:  the index of the channel.
            * str:  the name of the channel.
            * list of strings and integers:  a list of channels that will be
              used.
            * slice: a slice selecting the range of channels that will be
              used.
            * None: all the channels will be used.

        Returns
        -------
        float or array
            The resulting value. If more than one channe was selected the
            return object will be a 1D array containing the result of the
            procesing.
        """
        return self._applyFunctionTo(hjorthComplexity,i)

    def synchronizationLikelihood(self, channels = None, allPermutations=False,
                                  m=None, l=None, w1=None, w2=None, pRef=0.05,
                                  **kargs):
        """
        Returns the Synchronization Likelihood value applied over the i1 and i2
        channels by calling :func:`~eeglib.eeg.synchronizationLikelihood`.

        Parameters
        ----------
        channels: Variable type, optional
           In order to understand how this parameter works go to the doc of
           :py:meth:`eeglib.eeg.EEG._applyFunctionTo2C`
        allPermutations: bool, optional
            In order to understand how this parameter works go to the doc of
           :py:meth:`eeglib.eeg.EEG._applyFunctionTo2C`
        m: int, optional
            Numbers of elements of the embedded vectors.
        l: int, optional
            Separation between elements of the embedded vectors.
        w1: int, optional
            Theiler correction for autocorrelation effects
        w2: int, optional
            A window that sharpens the time resolution of the Synchronization
            measure
        pRef: float, optional
            The p Ref param of the synchronizationLikelihood. Default 0.05
        epsilonIterations: int,optional
            Number of iterations used to determine the value of epsilon

        Returns
        -------
        float or dict
            If a tuple is passed the it returns the result of applying the
            function to the channels specified in the tuple. If another valid
            value is passed, the method returns a dictionary, being the key the
            two channels used and the value the result of applying the function
            to those channels.
        """
        if l is None:
            l=1
        if m is None:
            m=int(np.sqrt(self.windowSize))
        l = 1 if l == 0 else l
        if w1 is None:
            w1 = int(2 * l * (m - 1))
        if w2 is None:
            w2 = int(10 // pRef + w1)
        return self._applyFunctionTo2C(lambda c1,c2:synchronizationLikelihood(
                                             c1,c2, m, l, w1, w2, pRef,**kargs)
                                        ,channels, allPermutations)

    def engagementLevel(self):
        """
        Returns the engagament level, which is calculated with this formula:
        beta/(alpha+theta), where alpha, beta and theta are the average of the
        average band values between al the channels.

        Returns
        -------
        float
            The engagement level.
        """
        bandValues = self.bandPower()
        alphas, betas, thetas = [], [], []
        for d in bandValues:
            alphas.append(d["alpha"])
            betas.append(d["beta"])
            thetas.append(d["theta"])

        alpha, beta, theta = np.mean(alphas), np.mean(betas), np.mean(thetas)

        return beta / (alpha + theta)

    def sampEn(self, i=None, *args, **kargs):
        """
        Returns Multiscale Sample Entropy at the given channel/s.

        Parameters
        ----------
        i: Variable type, optional
            * int:  the index of the channel.
            * str:  the name of the channel.
            * list of strings and integers:  a list of channels that will be
              used.
            * slice: a slice selecting the range of channels that will be
              used.
            * None: all the channels will be used
        m: int, optional
            Size of the embedded vectors. By default 2.
        l: int, optional
            Lag beetwen elements of embedded vectors. By default 1.
        r: float, optional
            Tolerance. By default fr*std(data)
        fr: float, optional
            Fraction of std(data) used as tolerance. If r is passed, this
            parameter is ignored. By default, 0.2.

        Returns
        -------
        float or array
            The resulting value. If more than one channe was selected the
            return object will be a 1D array containing the result of the
            procesing.
        """
        return self._applyFunctionTo(lambda x: sampEn(x,*args, **kargs), i)

    def LZC(self, i=None, *args, **kargs):
        """
        Returns the Lempel-Ziv Complexity at the given channel/s.

        Parameters
        ----------
        i: Variable type, optional
            * int:  the index of the channel.
            * str:  the name of the channel.
            * list of strings and integers:  a list of channels that will be
              used.
            * slice: a slice selecting the range of channels that will be
              used.
            * None: all the channels will be used
        theshold: numeric, optional
            A number use to binarize the signal. The values of the signal above
            threshold will be converted to 1 and the rest to 0. By default, the
            median of the data.
        normalize: bool
            If True the resulting value will be between 0 and 1, being 0 the
            minimal posible complexity of a sequence that has the same lenght
            of data and 1 the maximal posible complexity. By default, False.

        Returns
        -------
        float or array
            The resulting value. If more than one channe was selected the
            return object will be a 1D array containing the result of the
            procesing.
        """
        return self._applyFunctionTo(lambda x: LZC(x, *args, **kargs), i)

    def DFA(self, i=None, *args, **kargs):
        """
        Applies Detrended Fluctuation Analysis algorithm to the given data.

        Parameters
        ----------
         i: Variable type, optional
            * int:  the index of the channel.
            * str:  the name of the channel.
            * list of strings and integers:  a list of channels that will be
              used.
            * slice: a slice selecting the range of channels that will be
              used.
            * None: all the channels will be used
        fit_degree: int, optional
            Degree of the polynomial used to model de local trends. Default: 1.
        min_window_size: int, optional
            Size of the smallest window that will be used. Default:
            signalSize//2.
        fskip: float, optional
            Fraction of the window that will be skiped in each iteration for
            each window size. Default: 1.

        Returns
        -------
        float or array
            The resulting value. If more than one channe was selected the
            return object will be a 1D array containing the result of the
            procesing.
        """
        return self._applyFunctionTo(lambda x: DFA(x, *args, **kargs), i)

    def CCC(self, channels = None, allPermutations=False):
        """
        Computes the Cross Correlation Coeficient between the data in c1 and
        the data in c2.

        Parameters
        ----------
        channels: Variable type, optional
           In order to understand how this parameter works go to the doc of
           :py:meth:`eeglib.eeg.EEG._applyFunctionTo2C`
        allPermutations: bool, optional
            In order to understand how this parameter works go to the doc of
           :py:meth:`eeglib.eeg.EEG._applyFunctionTo2C`

        Returns
        -------
        float or dict
            If a tuple is passed the it returns the result of applying the
            function to the channels specified in the tuple. If another valid
            value is passed, the method returns a dictionary, being the key the
            two channels used and the value the result of applying the function
            to those channels.
        """
        f = lambda x,y: np.corrcoef(x,y)[0,1]
        return self._applyFunctionTo2C(f, channels,
                                        allPermutations=allPermutations)

    def DTW(self, channels=None, allPermutations = False, normalize = False,
            returnOnlyDistances = True ,*args, **kargs):
        """
        Computes the Dynamic Time Warping algortihm between the data of the
        given channels. It uses the FastDTW implementation given by the library
        fastdtw.

        Parameters
        ----------
        channels: Variable type, optional
           In order to understand how this parameter works go to the doc of
           :py:meth:`eeglib.eeg.EEG._applyFunctionTo2C`
        allPermutations: bool, optional
            In order to understand how this parameter works go to the doc of
           :py:meth:`eeglib.eeg.EEG._applyFunctionTo2C`
        normalize: bool optional
             If True the result of the algorithm is divided by the window size.
             Default: True.
        returnOnlyDistances: bool, optional
            If True, the result of the function will include only the distances
            after applying the DTW algorithm. If False it will return also the
            path. Default: True.
        radius : int, optional
            size of neighborhood when expanding the path. A higher value will
            increase the accuracy of the calculation but also increase time
            and memory consumption. A radius equal to the size of x and y will
            yield an exact dynamic time warping calculation.
        dist : function or int, optional
            The method for calculating the distance between x[i] and y[j]. If
            dist is an int of value p > 0, then the p-norm will be used. If
            dist is a function then dist(x[i], y[j]) will be used. If dist is
            None then abs(x[i] - y[j]) will be used.

        Returns
        -------
        tuple, float, dict of floats or dict of tuples
            If a tuple is passed the it returns the result of applying the
            function to the channels specified in the tuple. If another valid
            value is passed, the method returns a dictionary, being the key the
            two channels used and the value the result of applying the function
            to those channels.
        """

        fAux=lambda c1, c2: fastdtw.fastdtw(c1, c2, *args, **kargs)

        if normalize and returnOnlyDistances:
            f= lambda c1, c2: fAux(c1,c2)[0]/self.windowSize
        elif returnOnlyDistances:
            f= lambda c1, c2: fAux(c1,c2)[0]
        elif normalize:
            def f(c1,c2):
                ret = fAux(c1,c2)
                return ret[0]/self.windowSize,ret[1]
        else:
            f=fAux

        ret = self._applyFunctionTo2C(f, channels,
                                       allPermutations=allPermutations)

        return ret



"This module define the functions for preprocessing the signal data"


from scipy.signal import butter, filtfilt

def bandPassFilter(data,sampleRate=None,highpass=None,lowpass=None, order=2):
    """
    Return the signal filtered between highpass and lowpass. Note that neither
    highpass or lowpass should be above sampleRate/2.

    Parameters
    ----------
    data: numpy.ndarray
        The signal
    sampleRate: numeric, optional
        The frequency at which the signal was recorded. By default it is the
        same as the number of samples of the signal.
    highpass: numeric, optional
        The signal will be filtered above this value.
    lowpass: numeric, optional
        The signal will be filtered bellow this value.
    order: int, optional
        Butterworth

    Returns
    -------
    numpy.ndarray
        The signal filtered betwen the highpass and the lowpass
    """
    size=len(data)
    if not sampleRate:
        sampleRate=size

    #nyquist frequency
    nyq = 0.5*sampleRate

    if highpass:
        highpass=highpass/nyq

    if lowpass:
        lowpass=lowpass/nyq


    if lowpass and highpass:
        b,a = butter(order, [highpass, lowpass], btype="band")
    elif lowpass:
        b,a = butter(order, lowpass, btype="low")
    elif highpass:
        b,a = butter(order, highpass, btype="high")
    else:
        return data


    return filtfilt(b, a, data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module contains helper classes that are useful to iterating over a EEG
data stream. Currently there is support only for CSV files.
"""
import datetime

import csv
from pyedflib import EdfReader

import numpy as np
from scipy.stats import zscore
from sklearn.decomposition import FastICA

from eeglib.eeg import EEG
from eeglib.preprocessing import bandPassFilter


class Helper():
    """
    This is an abstract class that defines the way every helper works.
    """

    def __init__(self, data, sampleRate=None, windowSize=None, names=None,
                 highpass=None, lowpass=None, normalize=False, ICA=False,
                 selectedSignals=None):
        """
        Parameters
        ----------
        data: 2D matrix
            The signals data in the shape (nChannels, nSamples).
        sampleRate: numeric, optional
            The frequency at which the data was recorded. By default its value
            is the lenght of the data.
        windowSize: int, optional
            The size of the window in which the calculations will be done. By
            default its value is the lenght of one second of the data.
        names: list of strings
            A list containing the names of each channel in the same positions
            than data channels.
        highpass: numeric, optional
            The signal will be filtered above this value.
        lowpass: numeric, optional
            The signal will be filtered bellow this value.
        normalize: boolean, optional
            If True, the data will be normalizing using z-scores. Default =
            False.
        ICA: boolean, optional
            If True, Independent Component Analysis will be applied to the data
            . It is applied always after normalization if normalize = True.
            Default: False.
        selectedSignals: list of strings or ints
            If the data file has names asociated to each columns, those columns
            can be selected through the name or the index of the column. If the
            data file hasn't names in the columns, they can be selected just by
            the index.
        """

        self.data = data

        #Names check
        if names:
            self.names = names
        else:
            self.names = [str(i) for i in range(data.shape[0])]

        #Check selected signals
        if selectedSignals:
            self.selectSignals(selectedSignals)

        #SampleRate check
        if not sampleRate:
            self.sampleRate=len(self.data[0])
        else:
            self.sampleRate=sampleRate

        #windowSize check
        if not windowSize:
            self.windowSize = self.sampleRate
        else:
            self.windowSize = windowSize

        #Attributes inicialization
        self.nChannels = len(self.data)
        self.nSamples = len(self.data[0])
        self.startPoint = 0
        self.endPoint = self.nSamples
        self.step=None
        self.iterator = None
        self.duration = self.nSamples/self.sampleRate

        #Preparing inner EEG ovject
        self.prepareEEG(self.windowSize)

        #Lowpass and highpass check
        if lowpass or highpass:
            for i,channel in enumerate(self.data):
                self.data[i]=bandPassFilter(channel, self.sampleRate, highpass,
                                            lowpass)

        #ICA check
        if ICA:
            ica=FastICA()
            self.data=ica.fit_transform(self.data.transpose()).transpose()
            self.names = [str(i) for i in range(self.nChannels)]

        #normilze check
        if normalize:
            self.data=zscore(self.data,axis=1)

        #Setting an initial value for the EEG window
        self.moveEEGWindow(self.startPoint)

    def __iter__(self):
        return self._getIterator()

    def __len__(self):
        return self.nSamples

    def __getitem__(self, i):
        """
        Creates and return a ready iterator thet using the slice given as
        parameter.

        Parameters
        ----------
        i: slice
            The fields of the slice are used to create the iterator.

        Returns
        ----------
        Iterator
        """
        if not isinstance(i, slice):
            raise ValueError("Only slices can be used.")
        return self.prepareIterator(i.step, i.start, i.stop)

    def _getIterator(self):
        if not self.iterator:
            self.iterator = Iterator(self, self.step,
                                     self.startPoint,
                                     self.endPoint)

        return self.iterator

    def _handleIndex(self, index, step=False):
        #Is integer?
        if np.issubdtype(type(index), np.integer):
            if index<0 and not step:
                index = self.nSamples+index
        #Is string?
        elif isinstance(index, str):
            neg = index[0]=="-"

            if neg:
                index = index.replace("-","")

            index = index.split(":")
            nComponents = len(index)

            if 0 < nComponents < 4:
                totalSecs = 0
                for i in range(nComponents):
                    totalSecs+=float(index[-(i+1)])*60**i
                index = int(totalSecs*self.sampleRate)

                if neg:
                    index = self.nSamples-index

            else:
                raise ValueError("Wrong value for the index")
        #Is timedelta?
        elif isinstance(index, datetime.timedelta):
            index = int(self.sampleRate*index.total_seconds())

        else:
            raise ValueError("An index can only be a int, a str or a " +
                             "datetime.timedelta.")

        return index

    def prepareIterator(self, step=None, startPoint=0, endPoint=None):
        """
        Prepares the iterator of the helper.

        Parameters
        ----------
        step: int,optional
            Number of samples to be skipped in each iteration.
        startPoint: int, optional
            The index of first sample from where the iteration will start. By
            default 0.
        endPoint: int, optional
            The index of the last sample + 1 until where the iteration will go.
            By default the size of the data.
        """
        if startPoint:
            self.startPoint=self._handleIndex(startPoint)
        if endPoint:
            self.endPoint = self._handleIndex(endPoint)
        if step:
            self.step = self._handleIndex(step, step=True)

        self.iterator = Iterator(self,self.step,self.startPoint,self.endPoint)

        return self


    def prepareEEG(self, windowSize):
        """
        Prepares and creates the EEG object that the iteration will use with
        the same parameters that an EEG objects is initialized. Also it returns
        the inner eeg object.

        Parameters
        ----------
        windowSize: int
            The maximun samples the window will store.

        Returns
        -------
        EEG
        """
        self.windowSize = windowSize
        self.eeg = EEG(windowSize, self.sampleRate, self.nChannels,
                       names=self.names)
        self.windowPosition = None
        if not self.step:
            self.step = windowSize
        return self.eeg

    def moveEEGWindow(self, position):
        """
        Moves the window to start at position. Also it returns the inner eeg
        object.

        Parameters
        ----------
        position: int, str or datetime.timedelta
            * int: position of the sample
            * str with format hh:mm:ss.ss: temporal position
            * timedelta: temporal position

        Returns
        -------
        EEG
        """
        position =self._handleIndex(position)
        if position+self.eeg.windowSize>self.nSamples:
            raise ValueError("The start point is too near to the end.")

        self.eeg.set(self.data[:, position:position + self.eeg.windowSize]
                     ,columnMode=True)
        self.windowPosition = position
        return self.eeg


    def getNames(self, indexes=None):
        """
        Returns the names of the specified indexes of channels.

        Parameters
        ----------
        indexes: Iterable of int, optional.
            The indexes of the channels desired. If None it will return all the
            channels' names. Default: None.
        """
        if indexes:
            names = [self.names[i] for i in indexes]
        else:
            names = self.names

        return names

    def selectSignals(self, selectedSignals):
        """
        Parameters
        ----------
        selectedSignals : iterable of str and/or int
            The indexes of the desired selected signals. It must be an iterable
            type containing either str or int.

        Returns
        -------
        None.

        """
        for i, column in enumerate(selectedSignals):
            if isinstance(selectedSignals[i], str):
                selectedSignals[i]=self.names.index(column)

        self.names=[self.names[i] for i in selectedSignals]
        self.data=np.array([self.data[i] for i in selectedSignals])
        self.nChannels = len(self.names)

        if hasattr(self, "windowSize"):
            self.prepareEEG(self.windowSize)

class Iterator():
    def __init__(self, helper,step,auxPoint, endPoint):
        self.helper=helper
        self.step=step
        self.auxPoint=auxPoint
        self.endPoint=endPoint

    # Function for iterations
    def __next__(self):
        if self.auxPoint > self.endPoint-self.helper.eeg.windowSize:
            raise StopIteration
        self.helper.moveEEGWindow(self.auxPoint)
        self.auxPoint += self.step
        return self.helper.eeg

class CSVHelper(Helper):
    """
    This class is for applying diferents operations using the EEG class over a
    csv file.
    """
    def __init__(self, path, *args, **kargs):
        """
        The rest of parameters can be seen at :meth:`Helper.__init__`

        Parameters
        ----------
        path: str
            The path to the csv file
        """
        with open(path) as file:
            reader=csv.reader(file)
            l1=reader.__next__()
            data=[[] for _ in l1]
            for row in reader:
                for i, val in enumerate(row):
                    data[i].append(float(val))
        #Checking if first row is for channels names
        try:
            l1=list(map(float,l1))
            for value,column in zip(l1, data):
                column.insert(0,value)
            names = None
        except ValueError:
            names=l1

        data=np.array(data)

        if "names" in kargs:
            super().__init__(data, *args, **kargs)
        else:
            super().__init__(data, *args, names=names, **kargs)


class EDFHelper(Helper):
    """
    This class is for applying diferents operations using the EEG class over an
    edf file.
    """
    def __init__(self, path, *args, sampleRate=None, **kargs):
        """
        The rest of parameters can be seen at :meth:`Helper.__init__`

        Parameters
        ----------
        path: str
            The path to the edf file
        """
        reader = EdfReader(path)

        ns = reader.signals_in_file

        data  = [reader.readSignal(i) for i in range(ns)]
        names = reader.getSignalLabels()
        frequencies = reader.getSampleFrequencies()

        if not sampleRate:
            sampleRate = frequencies[0]
            if not all(frequencies==sampleRate):
                raise ValueError("All channels must have the same frequency.")

        data=np.array(data)

        super().__init__(data, *args, sampleRate = sampleRate, names = names,
             **kargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module is a library with some tools and functions for EEG signal analysis.

*Author*: Luis Cabañero Gómez
"""

from . import eeg
from . import helpers
from . import features
from . import preprocessing
from . import auxFunctions
from . import wrapper


"This module define the functions of the features for EEG analysis"

import numpy as np
import scipy as sp
from sklearn.neighbors import KDTree
from numba import njit

def bandPower(spectrum, bandsLimits, freqRes, normalize=False):
    """
        Returns the power of each band at the given index.

        Parameters
        ----------
        spectrum: 1D arraylike
            An array containing the spectrum of a signal

        bandsLimits: dict
            This parameter is used to indicate the bands that are going to be
            used. It is a dict with the name of each band as key and a tuple
            with the lower and upper bounds as value.

        freqRes: float
            Minimum resolution for the frequency.

        normalize: bool, optional
            If True the each band power is divided by the total power of the
            spectrum. Default False.

        Returns
        -------
        dict
            The keys are the name of each band and the values are their power.
        """
    total = 1
    if normalize:
        total = sp.integrate.trapz(spectrum, dx=freqRes)
    return {key:sp.integrate.trapz(spectrum[band[0]:band[1]], dx=freqRes)/total
            for key, band in bandsLimits.items()}

def PFD(data):
    """
    Returns the Petrosian Fractal Dimension of the signal given in data.

    Parameters
    ----------
    data: array_like
        Signal

    Returns
    -------
    float
        The resulting value
    """
    derivative = np.diff(data)
    size=len(data)
    signChanges = countSignChanges(derivative)
    logSize = np.log(size)
    return logSize / (logSize + np.log(size / (size + 0.4 * signChanges)))

def HFD(data,kMax=None):
    """
    Returns the Higuchi Fractal Dimension of the signal given data.

    Parameters
    ----------
    data: array_like
        signal

    kMax: int, optional
        By default it will be windowSize//4.

    Returns
    -------
    float
        The resulting value
    """
    #Inicializations
    data=np.array(data)
    N = len(data)
    kMax = N // 4 if kMax is None else kMax     #Default kMax
    L = np.zeros(kMax-1)
    x = np.array((-np.log(np.arange(2,kMax+1)),np.ones(kMax-1))).transpose()

    return _HFD(data, N, kMax, L, x)


@njit
def _HFD(data, N, kMax, L, x):# pragma: no cover
    # Loop from 2 to kMax
    for k in range(2, kMax + 1):
        Lk = np.zeros(k)
        #Loop for compute the lenght of Lmk
        for m in range(0, k):
            Lmk = 0
            for i in range(1, (N - m) // k):
                Lmk += abs(data[m + i * k] - data[m + i * k - k])
            Lk[m]=Lmk * (N - 1) / (((N - m) // k) * k * k)
        Laux=np.mean(Lk)
        Laux=0.01/k if Laux==0 else Laux
        L[k-2]=np.log(Laux)

    p, _, _, _ = np.linalg.lstsq(x, L)
    return p[0]

def synchronizationLikelihood(c1, c2, m, l, w1, w2, pRef=0.05, epsilonIterations=20):
    """
    Returns the Synchronization Likelihood between c1 and c2. This is a
    modified version of the original algorithm.

    Parameters
    ----------
    c1: array_like
        First signal
    c2: array_like
        second signal
    m: int
        Numbers of elements of the embedded vectors.
    l: int
        Separation between elements of the embedded vectors.
    w1: int
        Theiler correction for autocorrelation effects
    w2: int
        A window that sharpens the time resolution of the Synchronization
        measure
    pRef: float, optional
        The pRef param of the synchronizationLikelihood. Default 0.05
    epsilonIterations: int,optional
        Number of iterations used to determine the value of epsilon. Default:20

    Returns
    -------
    float
        A value between 0 and 1. 0 means that the signal are not synchronized
        at all and 1 means that they are totally synchronized.
    """
    if len(c1)!=len(c2):
        raise ValueError("c1 and c2 must have the same lenght")

    c1 = np.array(c1)
    c2 = np.array(c2)

    return _SL(c1, c2, m, l, w1, w2, pRef,epsilonIterations)

# Auxiliar functions for Synchronization Likeihood
@njit
def _SL(c1, c2, m, l, w1, w2, pRef, epsilonIterations):# pragma: no cover
    X1 = _getEmbeddedVectors(c1, m, l)
    X2 = _getEmbeddedVectors(c2, m, l)

    D1 = _getDistances(X1)
    D2 = _getDistances(X2)

    size=len(X1)
    E1 = np.zeros(size)
    E2 = np.zeros(size)
    for i in range(size):
        E1[i]=_getEpsilon(D1, i, pRef,epsilonIterations)
        E2[i]=_getEpsilon(D2, i, pRef,epsilonIterations)

    SL = 0
    SLMax = 0
    for i in range(size):
        Sij = 0
        SijMax = 0
        for j in range(size):
            if w1 < abs(j - i) < w2:
                if D1[i,j] < E1[i]:
                    if D2[i,j] < E2[i]:
                        Sij += 1
                    SijMax += 1
        SL += Sij
        SLMax += SijMax
    return SL / SLMax if SLMax>0 else 0

@njit
def _getHij(D, i, e):# pragma: no cover
    summ = 0
    for j in range(len(D)):
        if D[i,j] < e:
            summ += 1
    return summ

@njit
def _getDistances(X):# pragma: no cover
    t=len(X)
    D=np.zeros((t,t),dtype=np.float64)
    for i in range(t):
        for j in range(i):
            D[j,i]=D[i,j]=np.linalg.norm(X[i]-X[j])

    return D
@njit
def _getProbabilityP(D, i, e):# pragma: no cover
    return _getHij(D, i, e) /len(D)

@njit
def _getEmbeddedVectors(x, m, l):# pragma: no cover
    size = len(x) - (m - 1) * l
    X = np.zeros((m, size))
    for i in range(m):
        X[i]=x[i*l:i * l + size]

    return X.T

@njit
def _logDiference(p1,p2):# pragma: no cover
    return abs(np.log(p2/p1))

@njit
def _getEpsilon(D, i, pRef, iterations):# pragma: no cover
    eInf = 0
    eSup = None
    bestE=e = 1
    bestP=p = 1
    minP = 1 / len(D)
    for _ in range(iterations):
        p = _getProbabilityP(D, i, e)
        if pRef < minP == p:
            break
        
        if p < pRef:
            eInf = e
        elif p > pRef:
            eSup = e
        else:
            bestP=p
            bestE=e
            break
        
        if _logDiference(bestP,pRef) > _logDiference(p,pRef):
            bestP=p
            bestE=e
        e = e * 2 if eSup is None else (eInf + eSup) / 2

    return bestE


def countSignChanges(data):
    """
    Returns the number of sign changes of a 1D array

    Parameters
    ----------
    data: array_like
        The data from which the sign changes will be counted

    Returns
    -------
    int
        Number of sign changes in the data
    """
    signChanges = 0
    for i in range(1, len(data)):
        if data[i] * data[i - 1] < 0:
            signChanges += 1
    return signChanges


# HjorthParameters
def hjorthActivity(data):
    """
    Returns the Hjorth Activity of the given data

    Parameters
    ----------
    data: array_like

    Returns
    -------
    float
        The resulting value
    """
    return np.var(data)


def hjorthMobility(data):
    """
    Returns the Hjorth Mobility of the given data

    Parameters
    ----------
    data: array_like

    Returns
    -------
    float
        The resulting value
    """
    return np.sqrt(np.var(np.gradient(data)) / np.var(data))


def hjorthComplexity(data):
    """
    Returns the Hjorth Complexity of the given data

    Parameters
    ----------
    data: array_like

    Returns
    -------
    float
        The resulting value
    """
    return hjorthMobility(np.gradient(data)) / hjorthMobility(data)


# Sample Entropy
def sampEn(data, m = 2, l = 1, r = None, fr = 0.2, eps = 1e-10):
    """
    Returns Sample Entropy of the given data.

    Parameters
    ----------
    data: array_like
        The signal
    m: int, optional
        Size of the embedded vectors. By default 2.
    l: int, optional
        Lag beetwen elements of embedded vectors. By default 1.
    r: float, optional
        Tolerance. By default fr*std(data)
    fr: float, optional
        Fraction of std(data) used as tolerance. If r is passed, this
        parameter is ignored. By default, 0.2.
    eps: float, optional
        Small number added to avoid infinite results. If 0 infinite results can
        appear. Default: 1e-10.

    Returns
    -------
    float
        The resulting value
    """
    if not r:
        r = fr * np.std(data)

    A = _countEmbeddedDistances(data, m+1, l, r) + eps
    B = _countEmbeddedDistances(data, m  , l, r) + eps

    if B == 0:# pragma: no cover
        return -np.inf

    if A == 0:# pragma: no cover
        return np.inf
    return -np.log(A/B)

def _countEmbeddedDistances(data, m, l, r):
    X = _getEmbeddedVectors(data , m, l)

    kdtree = KDTree(X, metric="chebyshev")
    # Return the count
    return np.sum(kdtree.query_radius(X, r, count_only=True) - 1)


# Lempel-Ziv Complexity
def LZC(data, threshold = None):
    """
    Returns the Lempel-Ziv Complexity (LZ76) of the given data.

    Parameters
    ----------
    data: array_like
        The signal.
    theshold: numeric, optional
        A number use to binarize the signal. The values of the signal above
        threshold will be converted to 1 and the rest to 0. By default, the
        median of the data.

    References
    ----------
    .. [1] M. Aboy, R. Hornero, D. Abasolo and D. Alvarez, "Interpretation of
           the Lempel-Ziv Complexity Measure in the Context of Biomedical
           Signal Analysis," in IEEE Transactions on Biomedical Engineering,
           vol. 53, no.11, pp. 2282-2288, Nov. 2006.
    """
    if not threshold:
        threshold=np.median(data)

    n = len(data)

    sequence = _binarize(data, threshold)

    c = _LZC(sequence)
    b = n/np.log2(n)

    lzc = c/b

    return lzc

@njit
def _LZC(sequence):# pragma: no cover
    n = len(sequence)
    complexity = 1

    q0    = 1
    qSize = 1

    sqi   = 0
    where = 0

    while q0 + qSize <= n:
        # If we are checking the end of the sequence we just need to look at
        # the last element
        if sqi != q0-1:
            contained, where = _isSubsequenceContained(sequence[q0:q0+qSize],
                                                    sequence[sqi:q0+qSize-1])
        else:
            contained = sequence[q0+qSize] == sequence[q0+qSize-1]

         #If Q is contained in sq~, we increase the size of q
        if contained:
            qSize+=1
            sqi = where
        #If Q is not contained the complexity is increased by 1 and reset Q
        else:
            q0+=qSize
            qSize=1
            complexity+=1
            sqi=0

    return complexity


def _binarize(data, threshold):
    if  not isinstance(data, np.ndarray):
        data = np.array(data)

    return np.array(data > threshold, np.uint8)

@njit
def _isSubsequenceContained(subSequence, sequence):# pragma: no cover
    """
    Checks if the subSequence is into the sequence and returns a tuple that
    informs if the subsequence is into and where. Return examples: (True, 7),
    (False, -1).
    """
    n = len(sequence)
    m = len(subSequence)

    for i in range(n-m+1):
        equal = True
        for j in range(m):
            equal = subSequence[j] == sequence[i+j]
            if not equal:
                break

        if equal:
            return True, i

    return False, -1


# Detrended Fluctuation Analysis
def DFA(data, fit_degree = 1, min_window_size = 4, max_window_size = None,
        fskip = 1, max_n_windows_sizes=None):
    """
    Applies Detrended Fluctuation Analysis algorithm to the given data.

    Parameters
    ----------
    data: array_like
        The signal.
    fit_degree: int, optional
        Degree of the polynomial used to model de local trends. Default: 1.
    min_window_size: int, optional
        Size of the smallest window that will be used. Default: 4.
    max_window_size: int, optional
        Size of the biggest window that will be used. Default: signalSize//4
    fskip: float, optional
        Fraction of the window that will be skiped in each iteration for each
        window size. Default: 1
    max_n_windows_sizes: int, optional
        Maximum number of window sizes that will be used. The final number can
        be smaller once the repeated values are removed
        Default: log2(size)

    Returns
    -------
    float
        The resulting value
    """
    #Arguments handling
    data = np.array(data)

    size=len(data)

    if not max_window_size:
        max_window_size = size//4

    #Detrended data
    Y = np.cumsum(data - np.mean(data))

    #Windows sizes
    if not max_n_windows_sizes:
        max_n_windows_sizes = int(np.round(np.log2(size)))

    ns = np.unique(
          np.geomspace(min_window_size, max_window_size, max_n_windows_sizes,
                       dtype=int))

    #Fluctuations for each window size
    F = np.zeros(ns.size)

    #Loop for each window size
    for indexF,n in enumerate(ns):
        itskip = max(int(fskip * n),1)
        nWindows = int(np.ceil((size - n + 1) / itskip))

        #Aux x
        x = np.arange(n)

        y  = np.array([Y[i*itskip:i*itskip+n] for i in range(0,nWindows)])
        c  = np.polynomial.polynomial.polyfit(x, y.T, fit_degree)
        yn = np.polynomial.polynomial.polyval(x, c)

        F[indexF] = np.mean(np.sqrt(np.sum((y-yn)**2, axis=1)/n))

    alpha = np.polyfit(np.log(ns), np.log(F), 1)[0]

    if np.isnan(alpha): # pragma: no cover
        return 0

    return alpha



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open("readme.md", "r") as fh:
    long_description = fh.read()

__version__ = "0.4.1.1"

setup(name='eeglib',
      version = __version__,
      description='A library with some tools and functions for EEG signal analysis',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Luis Cabañero Gómez',
      author_email='Luis.Cabanero@uclm.es',
      url="https://github.com/Xiul109/eeglib",
      py_modules = ['eeglib'],
      license='MIT',
      classifiers=[
          'Development Status :: 4 - Beta',

          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',

          'License :: OSI Approved :: MIT License',

          'Programming Language :: Python :: 3.6',
          'Topic :: Scientific/Engineering',
      ],
      keywords='lib EEG signal analysis',

      packages=find_packages(exclude=["docs"]),

      install_requires = ['numpy','scipy',
                          'scikit-learn',
                          'numba',
                          'pandas',
                          'pyedflib',
                          'fastdtw',
                          'progressbar2'],

      test_require = ['colorednoise'],
      
      extras_require = {
          'doc': ["sphinx", "numpydoc"]
          }
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module contains the Wrapper class that wraps around a Helper and computes
every feature in each window configured. It also allows specifying labels for
the whole signal or for specific segments.
"""
import pandas as pd
import progressbar

from eeglib.auxFunctions import flatData
from eeglib.eeg import EEG

class _FeatureAdder():
    """
    This class is used by the wrapper to add features defined in the EEG object
    to the proccesing.
    """

    _excludedFunctions = ["set", "getBoundsForBand"]

    # Init
    def __init__(self, wrapper):
        #Setting attributes
        self._wrapper = wrapper
        self._eeg = wrapper.helper.eeg

        #Iterating over EEG.__dict__ items
        for key, value in EEG.__dict__.items():
            if (key[0] != "_" and callable(value) and
                not key in self._excludedFunctions):
                # If the conditions are fullfilled the modified function is
                # added to the __dict__
                self.__dict__[key] = self._adderCreator(value)

    # Creator of adder functions
    def _adderCreator(self, f):
        """
        Creates an adder method. When an adder method is called, it adds a
        feature, that is obtained by applying f to a signal, to the wrapper.

        Parameters
        ----------
        f: function
            The function

        Returns
        -------
        function
        """
        def func(*args, name=None, hideArgs=False, **kargs):
            hadName = name
            if not name:
                name = f.__name__

            if not hideArgs and not hadName:
                name+=str(args)+str(kargs)

            self._wrapper.functions[name]=(lambda: f(self._eeg, *args,
                                                     **kargs))

        func.__name__ = f.__name__
        func.__doc__ = """
        **It adds the next feature to the wrapper.**

        %s
        """%f.__doc__

        return func

    def __call__(self, functionName, *args, name=None, hideArgs=False,
                 **kargs):
        """
        Adds a feature that will be included in the dataset.

        Parameters
        ----------
        functionName: str
            The name of the function to call.
        name: str
            A custom name for the feature that will be visible in the df.
        hideArgs: bool
            If True, the arguments are not attached to the name in the df.
            Default: False.
        *args and **kargs:
            Parameters of the function that computes the feature.

        Returns
        -------
        None
        """
        f = getattr(EEG, functionName)

        self._adderCreator(f)(*args, name=name, hideArgs=hideArgs, **kargs)

class Wrapper():
    """
    This class wraps around a helper and allows getting multiple features at
    once. The main usage of this class if for generating features to use with
    machine learning techniques.
    """
    _indexName = "window_id"
    def __init__(self, helper, flat = True, flatSeparator = "_", store=True,
                 label = None, segmentation = None, onlySegments = False,
                 showProgress = False):
        """
        Parameters
        ----------
        helper: :py:class:`eeglib.helpers.Helper`
            The helper object that will be used to get the data and the
            iteration settings.
        flat: Bool
            If True the result of calling getFeatures() will be a one
            dimensional sequence of data. If False, the return value will be
            a list containing the result of each feature, that can be a float,
            a dict or an array. Default: True.
        separator: str
            It is used to separate the features names when flatten.
        store: Bool
            If True, the data will be stored in a self.storedFeatures. Default:
            True.
        label: object, optional
            This value will be added as a field in the data except if it is
            None. Default: None.
        segmentation: dict of tuples, optional
            Parameter format [((begin, end),label), ]; begin is the begin of
            the segment, end is the end of the segment and label the label
            realed to that segment. Begin and end can be either an int, a str
            or a datetime.timedelta, being the same that the parameter of
            :py:meth:`eeglib.helpers.Helper.moveEEGWindow`. The label of the
            unlabelled segments will be 0. If None, no segmentation will be
            added. Default: None.
        onlySegments: bool, optional
            If True, only windows between segments will be computed. Only used
            if a segmentation argument is used.
        showProgress: bool, optional
            If True, it will shown the progress when computing all the features.
        """
        self.helper = helper
        self.iterator = iter(self.helper)

        self.functions = {}

        self.flat = flat
        self.flatSeparator = flatSeparator

        self.store = store
        self.lastStored = -1
        if store:
            self.storedFeatures = []

        self.label = label
        if segmentation:
            self.segmentation = [((self.helper._handleIndex(s0),
                                   self.helper._handleIndex(se)), label)
                                 for (s0, se), label in segmentation]
            self.segmentation = sorted(self.segmentation)
            self.segmentationIndex = 0
            self.onlySegments = onlySegments
            self.insideSegment = False
        else:
            self.segmentation = None

        self.showProgress = showProgress
        self.addFeature = _FeatureAdder(self)

    def __iter__(self):
        self.iterator = iter(self.helper)
        # If show progress, init progress bar
        if self.showProgress:
            step = self.iterator.step
            w_size = self.helper.windowSize
            max_progress = (self.iterator.endPoint-w_size)/step+1
            self.progressBar = progressbar.ProgressBar(max_value=max_progress)
        return self

    def __next__(self):
        # Show progress
        if self.showProgress:
            progress = self.iterator.auxPoint/self.iterator.step
            self.progressBar.update(progress)
            
        # Move iterator
        next(self.iterator)
        
        # Manage segmentation and segmentation skips
        if self.segmentation:
            self._updateSegmentStatus()
            if self.onlySegments and not self.insideSegment:
                if self.segmentationIndex >= len(self.segmentation):
                    raise StopIteration
                segment = self.segmentation[self.segmentationIndex]
                self.iterator.auxPoint = segment[0][0]
                self.insideSegment = True
                next(self.iterator)
        return self.getFeatures()

    def _updateSegmentStatus(self):
        self.insideSegment = False
        # Loop through all segments to find the current one
        while self.segmentationIndex < len(self.segmentation):
            curSeg, _ = self.segmentation[self.segmentationIndex]
            curPoint = self.iterator.auxPoint - self.iterator.step
            # If current point inside bounds, segment found
            if curSeg[0] <= curPoint < curSeg[1]:
                self.insideSegment = True
                break
            # If current point is before the segment begin, we are out
            # the segment
            elif curPoint <curSeg[0]:
                break
            else:
                self.segmentationIndex+=1
                

    def addFeatures(self, features):
        """
        Parameters
        ----------
        features: list(tuple(str,list, dict))
            A list containing tuples that represent the parameters needed to
            add a single feature.
        """
        for f in features:
            if isinstance(f,tuple) and len(f) == 3:
                self.addFeature(f[0],*f[1],**f[2])
            elif isinstance(f, str):
                self.addFeature(f)
            else:
                raise ValueError("Only tuples of len 3 or str are valid values")

    def addCustomFeature(self, function, channels = None, twoChannels = False,
                         name = None, customArgs = [], customKwargs = {}):
        """
        Adds a custom feature that will be included in the dataset.

        Parameters
        ----------
        function: function
            The function to apply. It must take an array-like as first parameter 
            if twoChannels parameter is False or two array-like for each of the
            first two parameters if twoChannels is True.

        channels: Variable type, optional
            The channels over which the function will be applied.
            * int:  the index of the channel.
            * str:  the name of the channel.
            * list of strings and integers:  a list of channels.
            * slice:  a slice selecting the range of channels.
            * None: all the channels.
            The function to apply to the data to obtain the feature.

        twoChannels: bool
            If function receives two channels of data this parameter should be
            True. Default: False.
        
        name: str
            A custom name for the feature that will be visible in the df.
        
        customArgs: list
            A list of fixed arguments for the function.
        
        customKwargs: dict
            A dict of keyword arguments, where the key is the argument name and
            the value is the argument content.

        Returns
        -------
        None
        """

        if not name:
            name = function.__name__

        if not twoChannels:
            aux_func = lambda x: function(x, *customArgs, **customKwargs)
            f = lambda: self.helper.eeg._applyFunctionTo(aux_func, channels)
        else:
            aux_func = lambda x1, x2: function(x1, x2, *customArgs, **customKwargs)
            f = lambda: self.helper.eeg._applyFunctionTo2C(aux_func, channels)

        

        self.functions[name] = f


    def featuresNames(self):
        """
        Returns the names of the specified features.
        """
        return self.functions.keys()


    def getFeatures(self):
        """
        Returns the features in the current window of the helper iterator.

        Returns
        -------
        pandas.Series
        """
        if self.lastStored == self.iterator.auxPoint:
            features = self.storedFeatures[-1]
        else:
            features = {name:f() for name, f in self.functions.items()}
            next_window = self.iterator.auxPoint
            step = self.iterator.step
            features[self._indexName] = int(next_window/step - 1)
            if self.flat:
                features = flatData(features,"", self.flatSeparator)

            if self.segmentation:
                if self.insideSegment:
                    segment = self.segmentation[self.segmentationIndex]
                    features["segment_label"] = segment[1]
                else:
                    features["segment_label"] = 0
            
            if self.label is not None:
                features["label"] = self.label

            if self.store:
                self.storedFeatures.append(features)
                self.lastStored = self.iterator.auxPoint

        return pd.Series(features)

    def getStoredFeatures(self):
        """
        Returns the stored features if store was set to True, else it returns
        None.
        """
        if self.store:
            df = pd.DataFrame(self.storedFeatures)
            df.set_index(self._indexName, inplace=True)
            return df
        return None

    def getAllFeatures(self):
        """
        Iterates over all the windows in the helper and returns all the values.

        Returns
        -------
        pandas.DataFrame
        """
        data=[features for features in self]
        if self.store:
            data=self.storedFeatures
            
        df = pd.DataFrame(data)
        df.set_index(self._indexName, inplace=True)
        return df


    def reset(self):
        """
        Resets the

        Returns
        -------
        None.

        """
        self.iterator = iter(self.helper)



"""
This module includes auxiliar functions that are intended to be used only by
the library.
"""

def listType(l):
    """
    If the type of every element of the list l is the same, this function
    returns that type, else it returns None. If l is not a list, this function
    will raise a ValueError.
    """
    if not isinstance(l, list):
        raise ValueError("l is not a list.")

    if len(l) == 0:
        return None

    t = type(l[0])
    for e in l:
        if not isinstance(e, t):
            return None

    return t


def flatData(data, name, separator = "_"):
    """
    Transforms the parameter value into a flatten dictionary, using the
    parameter name as a prefix for the dictionary keys.

    """
    res={}

    auxName=name
    if name != "":
        auxName = auxName + separator

    if isinstance(data, dict):
        for key, val in data.items():
            res.update(flatData(val, auxName+str(key), separator))
    elif hasattr(data, "__iter__"):
        for i, val in enumerate(data):
            res.update(flatData(val, auxName+"%d"%i, separator))
    else:
        res = {name:data}

    return res
